/*
   Copyright (C) 2002-2019 CERN for the benefit of the ATLAS collaboration
   */

/**   @file ClusterReadAlg.cxx
 *   Implementation file for the ClusterReadAlg class.
 *   @author Dave Casper
 *   @date 28 February 2021
 */

#include "ClusterReadAlg.h"

#include "TrackerPrepRawData/FaserSCT_Cluster.h"
#include "FaserDetDescr/FaserDetectorID.h"    
#include "TrackerIdentifier/FaserSCT_ID.h"
#include "TrkEventPrimitives/FitQuality.h"
#include "TrkEventPrimitives/ParticleHypothesis.h"
#include "TrkEventPrimitives/LocalParameters.h"
#include "TrkRIO_OnTrack/RIO_OnTrack.h"
#include "StoreGate/ReadHandle.h"
#include <TH1F.h>
#include <TTree.h>
#include <TBranch.h>

/*
To add as new branches:
POOLCollectiontree - IsSimulation (0 or 1)
- event number
- event time
check issimulation before continuing
- coords, energy deposit, etc
*/


//static const std::string moduleFailureReason{"ClusterReadAlg: Exceeds max clusters"};

// Constructor with parameters:
ClusterReadAlg::ClusterReadAlg(const std::string& name, ISvcLocator* pSvcLocator) :
  AthReentrantAlgorithm(name, pSvcLocator), 
  AthHistogramming( name ),
  m_idHelper{nullptr},
  m_histSvc ( "THistSvc/THistSvc", name ){}

// Initialize method:
StatusCode ClusterReadAlg::initialize() {
  ATH_MSG_INFO("ClusterReadAlg::initialize()");
  ATH_CHECK(m_mcEventKey.initialize() );
  ATH_CHECK( m_faserSiHitKey.initialize() );
  //TH_CHECK( m_faserRdoKey.initialize());
  //ATH_CHECK( m_sctMap.initialize());
  ATH_CHECK(m_faserTriggerDataKey.initialize());
  ATH_CHECK(m_clusterContainerKey.initialize());
  ATH_CHECK(m_trackCollection.initialize());

  // Get the SCT ID helper
  ATH_CHECK(detStore()->retrieve(m_idHelper, "FaserSCT_ID"));

 // Histograms
 

  // Tree
  m_tree = new TTree("residTree","Cosmics residuals");
  m_tree->Branch("id", &m_hash, "id/I");
  m_tree->Branch("mc_Event", &m_mcEventKey, "mcEvent/F");
  m_tree->Branch("waveformExists", &m_waveformExists, "waveform/B");
  m_tree->Branch("station", &m_station, "station/I");

  //hist test
  m_hist = new TH1D("Stations", "Stations hit", 4, 0, 3);
  ATH_CHECK(histSvc()->regHist("/HIST/myhist", m_hist));
  m_hist2 = new TH1D("EnergyLoss", "Energy",100,0,0.1);
  ATH_CHECK(histSvc()->regHist("/HIST/Ehist", m_hist2));



  return StatusCode::SUCCESS;
}

// Execute method:
StatusCode ClusterReadAlg::execute(const EventContext& ctx) const
{


  ++m_numberOfEvents;  
  setFilterPassed(false, ctx);

  SG::WriteHandle trackContainer{m_trackCollection, ctx};
  std::unique_ptr<TrackCollection> outputTracks = std::make_unique<TrackCollection>();

  SG::ReadHandle<xAOD::FaserTriggerData> triggerContainer{m_faserTriggerDataKey, ctx};
  m_triggerisValid = triggerContainer.isValid();

  
  ++m_numberOfTriggeredEvents;

  SG::ReadHandle<Tracker::FaserSCT_ClusterContainer> clusterContainer{m_clusterContainerKey, ctx};
  m_clusterisValid = clusterContainer.isValid();

  // Loop over stations
  

  ATH_CHECK(trackContainer.record(std::move(outputTracks)));
  

  //RDO borrowed stuff
  // Handles created from handle keys behave like pointers to the corresponding container
  SG::ReadHandle<McEventCollection> h_mcEvents(m_mcEventKey);
  ATH_MSG_INFO("Read McEventContainer with " << h_mcEvents->size() << " events");
  ATH_MSG_INFO("\n\nTRIGGERS: Cluster:"<<m_clusterisValid<<"  Trigger:"<<m_triggerisValid);
  if (h_mcEvents->size() == 0) return StatusCode::FAILURE;
  if (m_clusterisValid || m_triggerisValid) {
    
     m_waveformExists = true;
  }

  if (m_clusterisValid && m_triggerisValid) {
     m_allTriggered = true;
     ATH_MSG_INFO("\n\nALL TRIGGERED\n\n"); 

  }

  // h_mcEvents->at(0)->print( msg().stream() );
  SG::ReadHandle<FaserSiHitCollection> h_siHits(m_faserSiHitKey);
  ATH_MSG_INFO("Read FaserSiHitCollection with " << h_siHits->size() << " hits");

  for (const FaserSiHit& hit : *h_siHits){
    int station= hit.getStation();
    int plane =hit.getPlane();
    int row =hit.getRow();
    int module = hit.getModule();
    int sensor = hit.getSensor();
    float energyloss= hit.energyLoss();
    //std::cout<<"St:"<<station<<"\tPl:"<<plane<<"\tMod:"<<module<<"\n";
    
    ++m_station[station];
    m_hist->Fill(station);
    if (energyloss!=0) {
      m_hist2->Fill(energyloss);
      std::cout<<"\nENERGY LOSS"<<energyloss<<'\n';
      }

  }
  //SG::ReadHandle<FaserSCT_RDO_Container> h_sctRDO(m_faserRdoKey);
  //SG::ReadHandle<TrackerSimDataCollection> h_collectionMap(m_sctMap);

  

  m_tree->Fill();

  // Done
  return StatusCode::SUCCESS;
}

// Finalize method:
StatusCode ClusterReadAlg::finalize() 
{
  ATH_MSG_INFO("ClusterReadAlg::finalize()");
  ATH_MSG_INFO( m_numberOfEvents << " events processed" );
  ATH_MSG_INFO( m_numberOfTriggeredEvents << " triggered events processed" );
  ATH_MSG_INFO( m_numberOfClusterCollection<< " cluster collections processed" );
  ATH_MSG_INFO( m_numberOfCluster<< " cluster processed" );

  ATH_MSG_INFO("station distribution" << m_station[0] <<" "<< m_station[1]<<" "<< m_station[2]<< m_station[3]);
  


  return StatusCode::SUCCESS;
}


